from dataclasses import dataclass
from abc import ABC, abstractmethod
import dagorama.api.api_pb2 as pb2

class RetryConfiguration(ABC):
    @abstractmethod
    def as_message(self) -> pb2.RetryMessage:
        """
        Map the retry configuration to a protobuf message

        """
        pass


@dataclass
class StaticRetry(RetryConfiguration):
    """
    Constant retry uses a static amount of seconds, up until
    the max attempt count.

    """
    max_attempts: int
    interval: int

    def as_message(self):
        return pb2.RetryMessage(
            maxAttempts=self.max_attempts,
            staticInterval=self.interval,
        )

@dataclass
class ExponentialRetry(RetryConfiguration):
    """
    Expontential retry uses the formula base_interval^t where t is the current
    attempt of the retry. When `max_attempts` has occurred will result in a permanent failure.

    """
    max_attempts: int
    base_interval: int = 2

    def as_message(self):
        return pb2.RetryMessage(
            maxAttempts=self.max_attempts,
            exponentialBase=self.base_interval,
        )



from dagorama.decorators import dagorama
from dagorama.definition import DAGDefinition, resolve
from dagorama.runner import execute, CodeMismatchException
import pytest
from contextlib import contextmanager


class CustomDag(DAGDefinition):
    """
    This DAG takes an input number and manipulates it arbitrarily.

    obj (1) --> obj2 (2) -->
                          obj3 (2) --> obj5 (4)
                                                 --> obj7 (9)
                          obj4 (3) --> obj6 (5)
    
    """
    @dagorama()
    def entrypoint(self, number: int):
        return self.linear_continue(number)

    @dagorama()
    def linear_continue(self, number: int):
        obj2 = number + 1
        results = [
            self.loop_map(obj2+i)
            for i in range(2)
        ]
        return self.loop_consolidate(results)

    @dagorama()
    def loop_map(self, number: int):
        # obj4 or obj5
        return number + 2

    @dagorama()
    def loop_consolidate(self, numbers: list[int]):
        # obj7
        return sum(numbers)


@contextmanager
def modify_entrypoint_signature():
    """
    Mock the DAG entrypoint with a new no-op function in order to change
    the code signature

    """
    old_name = CustomDag.entrypoint.original_fn.__name__
    CustomDag.entrypoint.original_fn.__name__ = "mock-name"
    yield
    CustomDag.entrypoint.original_fn.__name__ = old_name


def test_sample_dag(broker):
    dag = CustomDag()
    dag_result = dag(1)

    execute(infinite_loop=False)

    assert resolve(dag, dag_result) == 9


def test_sample_dag_worker_code_mismatch(broker):
    dag = CustomDag()

    # This will queue the entrypoint but not yet execute it
    dag(1)

    with modify_entrypoint_signature():
        with pytest.raises(CodeMismatchException):
            execute(infinite_loop=False)



import pytest
from dagorama_broker.launch import launch_broker

# Scope to the current unit test so we don't share results across tests
@pytest.fixture(scope="function")
def broker():
    with launch_broker():
        yield



from dataclasses import dataclass
from pickle import dumps, loads
from typing import Any

from dagorama.inspection import extract_promise_identifiers

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from dagorama.models.promise import DAGPromise


@dataclass
class DAGArguments:
    """
    Local python arguments that a DAG function should be run with. These only have to be serialized
    in the python runtime and therefore contain a full copy of DAGPromises or other local elements.

    To be queued in the DAG, these objects need to be pickle serializable.

    """
    # Can include static values and other DAGPromises
    # Technically these should not be "Any" but should be any object type that
    # can be pickled / json encoded over the wire
    # We should add a validation step to make sure this is true at call time
    calltime_args: list["DAGPromise" | Any]
    calltime_kwargs: dict[str, "DAGPromise" | Any]

    def to_server_bytes(self) -> bytes:
        # When serializing other DAGPromises, just include the identifiers since we don't need to
        # cache their parameters as well
        return dumps(
            DAGArguments(
                calltime_args=extract_promise_identifiers(self.calltime_args),
                calltime_kwargs=extract_promise_identifiers(self.calltime_kwargs),
            )
        )

    @classmethod
    def from_server_bytes(cls, b: bytes) -> "DAGArguments":
        return loads(b)



from dagorama.code_signature import calculate_function_hash
from multiprocessing import Pool


def test_code_signature_basic():
    def fn1():
        print("Testing")

    def fn2():
        print("Testing")

    def fn3():
        print("Different Value")

    # Full equality of name matters
    assert calculate_function_hash(fn1) != calculate_function_hash(fn2)

    # Normalize the names
    fn2.__name__ = "fn1"
    fn3.__name__ = "fn1"
    assert calculate_function_hash(fn1) == calculate_function_hash(fn2)
    assert calculate_function_hash(fn1) != calculate_function_hash(fn3)


def test_code_signature_dependencies():
    def fn1():
        print("Testing")

    # Should minimally be different
    assert calculate_function_hash(fn1) != calculate_function_hash(fn1, include_package_versions=True)


def get_hash(_):
    def fn1():
        print("Testing")

    return calculate_function_hash(fn1)


def test_same_hash_across_sessions():
    """
    Ensure that our hashing technique is resillient to separate process launches. This is explicitly
    not true with the built-in hash function:
    https://docs.python.org/3/reference/datamodel.html#object.__hash__
    http://www.ocert.org/advisories/ocert-2011-003.html
    """

    with Pool(10) as pool:
        results = pool.map(get_hash, range(100))
        assert len(set(results)) == 1



from dagorama.retry import RetryConfiguration, ExponentialRetry, StaticRetry
import pytest

@pytest.mark.parametrize(
    "policy",
    [
        StaticRetry(max_attempts=2, interval=1),
        ExponentialRetry(max_attempts=2, base_interval=2)
    ]
)
def test_convert_message(policy: RetryConfiguration):
    """
    Ensure we can serialize a policy into a protobuf message
    """
    policy.as_message()



from pickle import dumps, loads
from time import sleep
from contextlib import contextmanager

import grpc

import dagorama.api.api_pb2_grpc as pb2_grpc
import dagorama.api.api_pb2 as pb2
from dagorama.definition import dagorama_context
from dagorama.inspection import resolve_promises
from dagorama.models.arguments import DAGArguments
from dagorama.serializer import name_to_function
from multiprocessing import Process
from dagorama.code_signature import calculate_function_hash
from threading import Thread
from traceback import format_exc


class CodeMismatchException(Exception):
    """
    If raised, the queued DAGNode and the current runner are using
    different versions of the function code.

    """
    pass


def schedule_ping(
    context: pb2_grpc.DagoramaStub,
    worker: pb2.WorkerMessage,
    interval: int = 30,
):
    """
    Send a ping to the server to keep the connection alive.
    """
    while True:
        context.Ping(worker)
        sleep(interval)


def execute(
    exclude_queues: list | None = None,
    include_queues: list | None = None,
    queue_tolerations: list | None = None,
    infinite_loop: bool = True,
):
    with dagorama_context() as context:
        worker = context.CreateWorker(
            pb2.WorkerConfigurationMessage(
                excludeQueues=exclude_queues or [],
                includeQueues=include_queues or [],
                queueTolerations=queue_tolerations or [],
            )
        )

        Thread(
            target=schedule_ping,
            args=(context, worker),
            # Stop the thread when the rest of the process finishes
            daemon=True,
        ).start()

        while True:
            try:
                next_item = context.GetWork(worker)
            except grpc.RpcError as e:
                if e.details() == "no work available":
                    if not infinite_loop:
                        return
                    print("No work available, polling in 1s...")
                    sleep(1)
                    continue
                else:
                    raise

            arguments = DAGArguments.from_server_bytes(next_item.arguments)

            resolved_values = {
                source.identifier: loads(source.resolvedValue)
                for source in next_item.sources
            }

            # TODO: Add actual resolution
            print("------")
            print(next_item)
            print(arguments)
            print("resolved values", resolved_values)
            resolved_args = resolve_promises(arguments.calltime_args, resolved_values)
            resolved_kwargs = resolve_promises(arguments.calltime_kwargs, resolved_values)
            print("Resolved", resolved_args)
            print("Resolved", resolved_kwargs)

            resolved_fn = name_to_function(next_item.functionName, next_item.instanceId)

            # Ensure that we have the correct local version of the function
            print("FOUND FN", resolved_fn)
            print("COMPARE VALUES", calculate_function_hash(resolved_fn.original_fn), next_item.functionHash)
            if calculate_function_hash(resolved_fn.original_fn) != next_item.functionHash:
                raise CodeMismatchException()

            try:
                result = resolved_fn(*resolved_args, greedy_execution=True, **resolved_kwargs)
            except Exception as e:
                traceback = format_exc()
                context.SubmitFailure(
                    pb2.WorkFailedMessage(
                        instanceId=next_item.instanceId,
                        nodeId=next_item.identifier,
                        workerId=worker.identifier,
                        traceback=traceback
                    )
                )
                continue
            print("RESULT", result)

            context.SubmitWork(
                pb2.WorkCompleteMessage(
                    instanceId=next_item.instanceId,
                    nodeId=next_item.identifier,
                    workerId=worker.identifier,
                    result=dumps(result),
                )
            )


@contextmanager
def launch_workers(n: int = 1):
    """
    Helper function to spawn multiple workers without having to launch
    them in separate shell sessions.

    Mostly intended for unit testing.

    """
    workers = [Process(target=execute) for _ in range(n)]
    for worker in workers:
        worker.start()

    yield

    for worker in workers:
        worker.terminate()




from importlib import import_module
from inspect import getmodulename, isclass
from uuid import UUID

from dagorama.definition import DAGDefinition
from typing import cast


def function_to_name(func):
    module_name = func.__globals__["__name__"]
    class_path = func.__qualname__

    return f"{module_name}:{class_path}"


def name_to_function(name: str, instance_id: UUID):
    # Can also try to sniff for function names throughout entire global namespace, but this
    # is still conditioned on name conventions
    package_name, fn_path = name.split(":")
    fn_path_components = fn_path.split(".")

    try:
        mod = import_module(package_name)
    except ModuleNotFoundError:
        raise ValueError(f"Could not import function with path: {name}")

    for component in fn_path_components:
        if not hasattr(mod, component):
            raise ValueError(f"Unable to resolve dagorama function: {name}")

        mod = getattr(mod, component)

        if isclass(mod) and issubclass(mod, DAGDefinition):
            mod = mod() # type: ignore
            mod.instance_id = instance_id  # type: ignore
        elif isclass(mod):
            # Assume we can instantiate classes with no init arguments
            mod = mod()

    return mod



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dagorama/api/api.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x64\x61gorama/api/api.proto\x12\x04main\"d\n\x1aWorkerConfigurationMessage\x12\x15\n\rexcludeQueues\x18\x01 \x03(\t\x12\x15\n\rincludeQueues\x18\x02 \x03(\t\x12\x18\n\x10queueTolerations\x18\x03 \x03(\t\"#\n\rWorkerMessage\x12\x12\n\nidentifier\x18\x01 \x01(\t\"2\n\x1cInstanceConfigurationMessage\x12\x12\n\nidentifier\x18\x01 \x01(\t\"%\n\x0fInstanceMessage\x12\x12\n\nidentifier\x18\x01 \x01(\t\"\xdd\x01\n\x18NodeConfigurationMessage\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x14\n\x0c\x66unctionName\x18\x02 \x01(\t\x12\x14\n\x0c\x66unctionHash\x18\x03 \x01(\t\x12\x11\n\tqueueName\x18\x04 \x01(\t\x12\x11\n\ttaintName\x18\x05 \x01(\t\x12\x11\n\targuments\x18\x06 \x01(\x0c\x12\x11\n\tsourceIds\x18\x07 \x03(\t\x12\x12\n\ninstanceId\x18\x08 \x01(\t\x12!\n\x05retry\x18\t \x01(\x0b\x32\x12.main.RetryMessage\"\xe8\x01\n\x0bNodeMessage\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x14\n\x0c\x66unctionName\x18\x02 \x01(\t\x12\x14\n\x0c\x66unctionHash\x18\x03 \x01(\t\x12\x11\n\tqueueName\x18\x04 \x01(\t\x12\x11\n\ttaintName\x18\x05 \x01(\t\x12\x11\n\targuments\x18\x06 \x01(\x0c\x12\x15\n\rresolvedValue\x18\x07 \x01(\x0c\x12\"\n\x07sources\x18\x08 \x03(\x0b\x32\x11.main.NodeMessage\x12\x11\n\tcompleted\x18\t \x01(\x08\x12\x12\n\ninstanceId\x18\n \x01(\t\"T\n\x0cRetryMessage\x12\x13\n\x0bmaxAttempts\x18\x01 \x01(\x05\x12\x16\n\x0estaticInterval\x18\x02 \x01(\x05\x12\x17\n\x0f\x65xponentialBase\x18\x03 \x01(\x05\"\x1f\n\x0bPongMessage\x12\x10\n\x08lastPing\x18\x02 \x01(\x03\"[\n\x13WorkCompleteMessage\x12\x12\n\ninstanceId\x18\x01 \x01(\t\x12\x0e\n\x06nodeId\x18\x02 \x01(\t\x12\x10\n\x08workerId\x18\x03 \x01(\t\x12\x0e\n\x06result\x18\x04 \x01(\x0c\"\\\n\x11WorkFailedMessage\x12\x12\n\ninstanceId\x18\x01 \x01(\t\x12\x0e\n\x06nodeId\x18\x02 \x01(\t\x12\x10\n\x08workerId\x18\x03 \x01(\t\x12\x11\n\ttraceback\x18\x04 \x01(\t\"=\n\x13NodeRetrieveMessage\x12\x12\n\ninstanceId\x18\x01 \x01(\t\x12\x12\n\nidentifier\x18\x02 \x01(\t2\x84\x04\n\x08\x44\x61gorama\x12G\n\x0c\x43reateWorker\x12 .main.WorkerConfigurationMessage\x1a\x13.main.WorkerMessage\"\x00\x12M\n\x0e\x43reateInstance\x12\".main.InstanceConfigurationMessage\x1a\x15.main.InstanceMessage\"\x00\x12\x41\n\nCreateNode\x12\x1e.main.NodeConfigurationMessage\x1a\x11.main.NodeMessage\"\x00\x12\x30\n\x04Ping\x12\x13.main.WorkerMessage\x1a\x11.main.PongMessage\"\x00\x12\x39\n\x07GetNode\x12\x19.main.NodeRetrieveMessage\x1a\x11.main.NodeMessage\"\x00\x12\x33\n\x07GetWork\x12\x13.main.WorkerMessage\x1a\x11.main.NodeMessage\"\x00\x12<\n\nSubmitWork\x12\x19.main.WorkCompleteMessage\x1a\x11.main.NodeMessage\"\x00\x12=\n\rSubmitFailure\x12\x17.main.WorkFailedMessage\x1a\x11.main.NodeMessage\"\x00\x42\x0eZ\x0c\x64\x61gorama/apib\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dagorama.api.api_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\014dagorama/api'
  _WORKERCONFIGURATIONMESSAGE._serialized_start=32
  _WORKERCONFIGURATIONMESSAGE._serialized_end=132
  _WORKERMESSAGE._serialized_start=134
  _WORKERMESSAGE._serialized_end=169
  _INSTANCECONFIGURATIONMESSAGE._serialized_start=171
  _INSTANCECONFIGURATIONMESSAGE._serialized_end=221
  _INSTANCEMESSAGE._serialized_start=223
  _INSTANCEMESSAGE._serialized_end=260
  _NODECONFIGURATIONMESSAGE._serialized_start=263
  _NODECONFIGURATIONMESSAGE._serialized_end=484
  _NODEMESSAGE._serialized_start=487
  _NODEMESSAGE._serialized_end=719
  _RETRYMESSAGE._serialized_start=721
  _RETRYMESSAGE._serialized_end=805
  _PONGMESSAGE._serialized_start=807
  _PONGMESSAGE._serialized_end=838
  _WORKCOMPLETEMESSAGE._serialized_start=840
  _WORKCOMPLETEMESSAGE._serialized_end=931
  _WORKFAILEDMESSAGE._serialized_start=933
  _WORKFAILEDMESSAGE._serialized_end=1025
  _NODERETRIEVEMESSAGE._serialized_start=1027
  _NODERETRIEVEMESSAGE._serialized_end=1088
  _DAGORAMA._serialized_start=1091
  _DAGORAMA._serialized_end=1607
# @@protoc_insertion_point(module_scope)



from click import command, option
from dagorama.runner import execute

@command()
@option("--exclude-queue", multiple=True)
@option("--include-queue", multiple=True)
@option("--toleration", multiple=True)
def worker(exclude_queue: list[str], include_queue: list[str], toleration: list[str]):
    execute(
        exclude_queues=exclude_queue,
        include_queues=include_queue,
        queue_tolerations=toleration,
    )



from dagorama.decorators import dagorama
from dagorama.definition import DAGDefinition, resolve
from dagorama.runner import execute
from dagorama.retry import StaticRetry, ExponentialRetry
import pytest


class CustomNameDag(DAGDefinition):
    @dagorama(queue_name="test_queue")
    def entrypoint(self):
        return 10


class CustomTaintDag(DAGDefinition):
    @dagorama(taint_name="test_taint")
    def entrypoint(self):
        return 10


class CustomStaticErroringDag(DAGDefinition):
    @dagorama(retry=StaticRetry(max_attempts=2, interval=1))
    def entrypoint(self):
        raise ValueError()


class CustomExponentialErroringDag(DAGDefinition):
    @dagorama(retry=ExponentialRetry(max_attempts=2, base_interval=2))
    def entrypoint(self):
        raise ValueError()


def test_custom_name(broker):
    dag = CustomNameDag()
    dag_result = dag()

    # Should run this queue
    execute(
        include_queues=["test_queue"],
        infinite_loop=False
    )
    assert resolve(dag, dag_result) == 10

    dag = CustomNameDag()
    dag_result = dag()

    # Should not run this queue
    execute(
        include_queues=["test_queue_2"],
        infinite_loop=False
    )
    assert resolve(dag, dag_result) == None

def test_taint_name(broker):
    dag = CustomTaintDag()
    dag_result = dag()

    # Should not run a tained queue by default
    execute(infinite_loop=False)
    assert resolve(dag, dag_result) == None

    # Require specific allowance
    execute(queue_tolerations=["test_taint"], infinite_loop=False)
    assert resolve(dag, dag_result) == 10


@pytest.mark.parametrize("dag_class", [CustomStaticErroringDag, CustomExponentialErroringDag])
def test_erroring_dags(broker, dag_class):
    dag = dag_class()
    dag_result = dag()

    # Should not run a tained queue by default
    execute(infinite_loop=False)
    assert resolve(dag, dag_result) == None



from typing import Any, Callable

from dagorama.models.promise import DAGPromise


def find_promises(obj: Any) -> list[DAGPromise]:
    """
    Find all DAGPromises that are explicitly within the object, given standard python types. Note that this
    is just intended to find one layer of dependencies from obj. It does not recurse over additional dependencies
    that might be part of the found promises.

    """
    found_promises = []

    def find_function(promise: DAGPromise) -> None:
        nonlocal found_promises
        found_promises.append(promise)
        return None

    map_promises(obj, find_function)
    return found_promises


def resolve_promises(obj: Any, resolved_values: dict[str, Any]) -> Any:
    """
    Recursively resolves promises in the given structures they appear

    """
    def resolution_function(promise: DAGPromise) -> Any:
        # We should have been provided a value for this dependency from the server
        if str(promise.identifier) not in resolved_values:
            raise ValueError("Unexpected promise in resolved value")
        return resolved_values[str(promise.identifier)]

    return map_promises(obj, resolution_function)


def extract_promise_identifiers(obj: Any) -> Any:
    """
    Instead of persisting full DAGPromise objects, we just want to persist the identifier. This
    identifier should already be mapped to a fully hydrated version of the promise in the server.

    """
    def simplify_function(promise: DAGPromise) -> DAGPromise:
        return DAGPromise(identifier=promise.identifier)

    return map_promises(obj, simplify_function)


def map_promises(obj: Any, identifier_mapper: Callable[[DAGPromise], Any]) -> Any:
    """
    Recursively map promises given a user-defined function. Used by other functions that want
    to iterative build up some logic in the full argument chain.

    Supports nested lists and dictionaries

    """
    # Deal with iterables
    if isinstance(obj, dict):
        return {
            key: map_promises(value, identifier_mapper)
            for key, value in obj.items()
        }
    elif isinstance(obj, set):
        return {
            map_promises(value, identifier_mapper)
            for value in obj
        }
    elif isinstance(obj, list):
        return [
            map_promises(value, identifier_mapper)
            for value in obj
        ]
    elif isinstance(obj, tuple):
        return tuple([
            map_promises(value, identifier_mapper)
            for value in obj
        ])

    # Otherwise we just have a raw value
    if isinstance(obj, DAGPromise):
        return identifier_mapper(obj)
    else:
        return obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dagorama', 'dagorama.api', 'dagorama.models', 'dagorama.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0', 'grpcio>=1.50.0,<2.0.0']

entry_points = \
{'console_scripts': ['worker = dagorama.cli:worker']}

setup_kwargs = {
    'name': 'dagorama',
    'version': '0.1.0',
    'description': '',
    'long_description': '# dagorama (WIP)\n\nDagorama is an opinionated computation library for use in background processing. You can think of it as a hybrid between Celery, Airflow, and Dask.\n\nIts primary design goal is to let users write chained data processing logic in vanilla Python and easily scale to external dedicated machines. In addition, it strives to provide:\n- Easily interpretable control flow in vanilla Python, without having to learn a new language ORM\n- Be written for modern Python, with support for IDE suggestions and mypy typehints\n- Support dynamic logic, like branching if statements that condition a sub-tree execution on some runtime conditions\n- Easily integrate with unit tests on one machine when testing\n- Provide sensible defaults for background processing\n\n## Quick Start\n\nLet\'s take this trivial example, representative of wanting to perform a series of long running blocking tasks. Those tasks could be ML inference, image rendering, or data aggregation. In any case it needs to process for a few seconds before returning results. In particular we care about logic that fully utilizes the compute resources of the current process. This is unlike web requests, which have network induced latency and can be trivially parallelized through asyncio.\n\n```python\nfrom collections import Counter\nfrom time import sleep, time\n\nclass MyTask:\n    def entrypoint(self):\n        results = [\n            self.perform_work(i)\n            for i in range(4)\n        ]\n\n        return self.rollup_statuses(results)\n\n    def perform_work(self, identifier: int) -> int:\n        sleep(2)\n        return identifier // 2\n\n    def rollup_statuses(self, responses: list[int]):\n        return Counter([status for status in responses])\n\ndef main():\n    start = time()\n    task = MyTask()\n    response = task.entrypoint()\n    end = time()\n    print(f"Response: {response} in {round(end-start, 2)}s")\n    assert response == Counter({0: 2, 1: 2})\n\nif __name__ == "__main__":\n    main()\n```\n\nIf you let this run, you should see an echoed counter after a bit of processing.\n\n```bash\n$ python samples/test1.py\nResponse: Counter({0: 2, 1: 2}) in 8.01s\n```\n\nThis code also has valid typehints, so mypy is happy during the typehinting.\n\n```bash\n$ mypy samples/test1.py\nSuccess: no issues found in 1 source file\n```\n\nNaturally we want to get our results as quickly as possible, while scaling to the available resources of the machine (or machines) that are in our compute cluster. This would _almost_ be a natural fit for something like a Celery queue - we\'ll spawn four separate jobs, one for each `perform_work` function. But how do we handle the aggregation stage? Do we block on a main process until it\'s done? How is state managed? What if the main process exits before we\'re completed, how can we pick up from where we were before?\n\nThis dependency chaining actually isn\'t great for queues. Instead you\'ll want something more akin to a computation graph or DAG, one that can condition later functions on the successful completion of previous functions. Here\'s how you would write the same thing in dagorama.\n\n```python\nfrom dagorama.decorators import dagorama\nfrom dagorama.definition import DAGDefinition, resolve\nfrom dagorama.runner import launch_workers\nfrom dagorama_broker.launch import launch_broker\n\nfrom collections import Counter\nfrom time import sleep, time\n\nclass MyTask(DAGDefinition):\n    @dagorama()\n    def entrypoint(self):\n        results = [\n            self.perform_work(i)\n            for i in range(4)\n        ]\n\n        return self.rollup_statuses(results)\n\n    @dagorama()\n    def perform_work(self, identifier: int) -> int:\n        sleep(2)\n        return identifier // 2\n\n    @dagorama()\n    def rollup_statuses(self, responses: list[int]):\n        return Counter([status for status in responses])\n\ndef main():\n    with launch_broker():\n        start = time()\n        task = MyTask()\n        promise = task()\n\n        with launch_workers(4):\n            sleep(3)\n        response = resolve(task, promise)\n        end = time() \n\n    print(f"Response: {response} in {round(end-start, 2)}s")\n    assert response == Counter({0: 2, 1: 2})\n\nif __name__ == "__main__":\n    main()\n```\n\nFor the sake of fitting this logic in one script there are a few different things going on here.\n\n1. We\'ve wrapped each function in a `@dagorama` decorator. This decorator indicates that a function execution like `self.perform_work()` or `self.rollup_statuses()` should be performed on a separate worker node. This is akin to launching a new task within a conventional queue.\n2. We launch the background broker with `with launch_broker()`. This will spawn a separate broker process that coordinates across multiple workers.\n3. We launch the workers themselves with launch_workers. In this case we perform the work in separate processes. This could just as easily be on separate machines without changing the APIs or code.\n\nUnlike before, we now complete in roughly the time for the primary work.\n\n```\n$ python run samples/test2.py\nResponse: Counter({0: 2, 1: 2}) in 3.03s\n```\n\nMypy is similarly happy with our DAG definition.\n```bash\n$ mypy samples/test2.py\nSuccess: no issues found in 1 source file\n```\n\nYou\'ll notice the diff of the core MyTask class is very small:\n\n```bash\n$ diff samples/test1.py samples/test2.py \n1c1,2\n< class MyTask:\n---\n> class MyTask(DAGDefinition):\n>     @dagorama()\n9a11\n>     @dagorama()\n13a16\n>     @dagorama()\n```\n\nThis is the core design goal of dagorama: write vanilla python and scale easily.\n\n## API Notes\n\nThis section attempts to be the only section you\'ll need to know to use dagorama in day-to-day development, without starting to offroad.\n\nEach group of logical code that you want to flow to one another should be contained in a class that inherits from `DAGDefinition`. You\'ll want this code to be deployed on each worker node so they\'re able to access the same core definition files. Docker is the preferred mechanism to ensure that the same code is mirrored on each device and computations will happen as expected. Each instance of a DAGDefinition created via `DAGDefinition()` wraps one execution of the functions within the DAG. If you\'re processing 5 separate input images, for example, you\'ll want to spawn 5 DAGDefinitions.\n\nThe dagorama broker will ensure that earlier DAG instances will complete before ones that are invoked later. The prioritization scheme is FIFO on the DAG instantiation order. This is useful in situations where you want to decrease the latency from start of processing to DAG completion for use in near-realtime logic.\n\nEach function that you want to execute on a separate machine should be wrapped in a `@dagorama` decorator. Calls to this function will be added to the computational graph and distributed appropriately. Class functions that aren\'t decorated will be run inline to the current executor.\n\nA `@dagorama` decorated function will _look_ like it returns the typehinted values to static analyzers like mypy. This allows you to write more interpretable code by passing around logical values. In reality, however, @dagorama functions will return a `DAGPromise` at runtime. This DAGPromise is meaningless - it doesn\'t have a value yet, since it hasn\'t yet been passed to a runner. These response values should only be passed to other `@dagorama` decorated functions as function arguments. When this is done workers will only execute that function once all its dependencies have been realized.\n\nThere are some situations where you want to limit the functions that will run on certain hardware. A common case is for ML functions to only be run on GPU accelerated devices. We follow the kubernetes model here by adding a taint to each function that shouldn\'t be deployed by default, like `@dagorama(taint_name="GPU")`. To pull from this queue workers will have to explicitly specify this taint, otherwise they won\'t pull from the backing queue.\n\nTo launch a worker function, install the python package in your virtualenv and run:\n\n```\nworker [--include-queue {queue}] [--exclude-queue {queue}] [--toleration {toleration}]\n```\n\n## Production Deployment (WIP)\n\nOutside of local testing, you probably won\'t want to run the workers on your same machine. Instead you\'ll want to distribute them across multiple machines. In this setup we recommend:\n\n- 1 Broker: Go executable running on a VM, kubernetes pod, or within docker. Backed by a persistent database to resume state in case of runtime interruptions.\n- N Workers. The devices that perform the computation. Can be the same physical hardware configurations or different depending on usecase.\n- M Spawners. The service that will first instantiate the DAG. This is typically the end application like a server backend.\n\n## Typehinting\n\nDagorama should satisfy typehints the same way that normal functions do. In other words, you can treat the DAGPromise returned as a fulfilled value before passing it into other functions downstream of the main DAG.\n\n## Development\n\nHacking on dagorama is encouraged. Here are some quick getting started steps.\n\n### Installing Python Dependencies\n\nWe manage our dependencies with poetry. It\'s not strictly speaking necessary (the pyproject.toml should install via pip in a standard virtual environment) but.\n\nIf you don\'t already have Poetry, install it [here](https://python-poetry.org/docs/).\n\n```\npoetry install\n```\n\n### Installing gRPC\n\nClients communicate with the broker over gRPC. You\'ll need support to generate the protobuf files within Golang and Python.\n\nGolang quick start: https://grpc.io/docs/languages/go/quickstart/\n\n```\ngo install google.golang.org/protobuf/cmd/protoc-gen-go@v1.28\ngo install google.golang.org/grpc/cmd/protoc-gen-go-grpc@v1.2\nexport PATH="$PATH:$(go env GOPATH)/bin"\n```\n\nWhen you update the grpc files, re-generate the client and server definition files via:\n\n```\n./build_protobuf.sh\n```\n\n### Unit Tests\n\nIf you want to run unit tests you\'ll also need `dagorama-broker` installed. This convenience package allows the tests to dynamically spawn and tear down a broker via pytest fixtures.\n\n```\npoetry run pip install -e ./dagorama-broker\n```\n',
    'author': 'Pierce Freeman',
    'author_email': 'pierce@freeman.vc',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from contextlib import contextmanager
from pickle import loads
from typing import cast
from uuid import UUID, uuid4

import grpc

import dagorama.api.api_pb2 as pb2
import dagorama.api.api_pb2_grpc as pb2_grpc
from dagorama.models.promise import DAGPromise

RUN_LOOP_PROMISES: list[DAGPromise] = []


@contextmanager
def dagorama_context():
    # TODO: Get global context otherwise creates it

    with grpc.insecure_channel("localhost:50051") as channel:
        yield pb2_grpc.DagoramaStub(channel)


class DAGDefinition(ABC):
    def __init__(self):
        self.instance_id : UUID | None = None

    def __call__(self, *args, **kwargs) -> DAGPromise:
        if self.instance_id is not None:
            raise ValueError("Can only spawn one instance of a DAGDefinition")

        # Calling indicates that we should spin off a new DAG instance
        with dagorama_context() as context:
            self.instance_id = uuid4()
            context.CreateInstance(
                pb2.InstanceConfigurationMessage(
                    identifier=str(self.instance_id)
                )
            )

        result_promise = self.entrypoint(*args, **kwargs)
        # Cast as an actual result promise since this is what clients expect
        return cast(DAGPromise, result_promise)

    @abstractmethod
    def entrypoint(self, *args, **kwargs):
        pass


def resolve(dag: DAGDefinition, promise: DAGPromise):
    """
    Given a promise (typically of the initial_entrypoint), will recursively resolve
    promises in a return value chain.

    ie. Promise A -> Promise B -> Promise C will shortcut to Promise C, which will
    then return the true value.

    """
    # TODO: Remove the DAGDefinition, we should be able to infer
    # the instance ID from the promise

    with dagorama_context() as context:
        current_return_value = promise

        while isinstance(current_return_value, DAGPromise):
            node = context.GetNode(
                pb2.NodeRetrieveMessage(
                    instanceId=str(dag.instance_id),
                    identifier=str(current_return_value.identifier),
                )
            )

            if not len(node.resolvedValue):
                return None

            resolved = loads(node.resolvedValue)
            if resolved is None:
                return None

            current_return_value = resolved

        return current_return_value



from collections.abc import Callable
from typing import ParamSpec, TypeVar, cast
from uuid import uuid4
from functools import wraps

import dagorama.api.api_pb2 as pb2
from dagorama.definition import DAGDefinition, dagorama_context
from dagorama.inspection import find_promises
from dagorama.models.arguments import DAGArguments
from dagorama.models.promise import DAGPromise
from dagorama.serializer import function_to_name
from dagorama.code_signature import calculate_function_hash
from dagorama.retry import RetryConfiguration

T = TypeVar('T')
P = ParamSpec('P')


def dagorama(
    queue_name: str | None = None,
    taint_name: str | None = None,
    retry: RetryConfiguration | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    The actual return type of functions wrapped with @dagorama() will be a DAGPromise. This is not what we want during
    development because this is a typeless type. Instead, we want to program the graph as if all promises are instantly
    fulfilled and the values are passed downstream. A decorator gives us this behavior because we're only performing the
    wrap at runtime, so the type hinting will correctly recommend the full-fledged type.

    :param taint: When a taint is provided, workers will need to explicitly provide a toleration to execute. This is used
        in cases where there are specific compute resources that should own one phase of the pipeline.

    """
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if not isinstance(args[0], DAGDefinition):
                raise ValueError("@dagorama can only wrap class methods")

            dag_definition : DAGDefinition = args[0]

            if dag_definition.instance_id is None:
                raise ValueError("DAGDefinition must be instantiated with call() before calling a method")

            # Can't be provided as an explicit keyword parameter because of a mypy constraint with P.kwargs having
            # to capture everything
            # https://github.com/python/typing/discussions/1191
            greedy_execution = kwargs.pop("greedy_execution", False)
            if greedy_execution:
                return func(*args, **kwargs)

            # Strip out the class definition when we store the arguments
            isolated_args = cast(list, args)[1:]

            # This function will have a result
            # Queue in the DAG backend
            promise = DAGPromise(
                uuid4(),
                function_to_name(func),
                DAGArguments(
                    isolated_args,
                    kwargs
                )
            )

            # Find the dependencies
            promise_dependencies = find_promises([isolated_args, kwargs])

            # Add to the remote runloop
            with dagorama_context() as context:
                context.CreateNode(
                    pb2.NodeConfigurationMessage(
                        identifier=str(promise.identifier),
                        functionName=cast(str, promise.function_name),
                        functionHash=calculate_function_hash(func),
                        taintName=taint_name or "",
                        queueName=queue_name or cast(str, promise.function_name),
                        arguments=(
                            cast(
                                # We know this is a valid argument object because we just set it
                                DAGArguments,
                                promise.arguments,
                            )
                            .to_server_bytes()
                        ),
                        sourceIds=[
                            str(dependency.identifier)
                            for dependency in promise_dependencies
                        ],
                        instanceId=str(dag_definition.instance_id),
                        retry=retry.as_message() if retry is not None else None,
                    )
                )

            return cast(
                # Wrong cast of types but we want the static typechecker to believe that the function
                # is returning the actual value as specified by the client caller
                # https://docs.python.org/3/library/typing.html#typing.ParamSpec
                T, promise,
            )
        wrapper.original_fn = func  # type: ignore
        return wrapper
    return decorator
